/*
** etk_drv.c
**
** Implements a subset of tcl to execute erlang
** functions (as scripts)
**
**  Works with: Tk4.2 (& Tcl7.6) 
**  solaris threads/ pthreads / win32 
**
*/

#include <stdio.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef __WIN32__
#define INCL_WINSOCK_API_TYPEDEFS 1
#include <winsock2.h>
#include <windows.h>
#include <process.h>
#else
#include <signal.h>
#endif

#include <tk.h>      /* correct version of tk.h */

#include "driver.h"

#include "etk.h"
#include "tkstr.h"


#if (TK_MAJOR_VERSION != 4) && (TK_MINOR_VERSION != 2)
#error "bad tk version should be 4.2"
#endif

#if (TCL_MAJOR_VERSION != 7) && (TCL_MINOR_VERSION != 6)
#error "bad tcl version should be 7.6"
#endif

#ifdef __WIN32__

typedef struct {
  LPFN_WSASTARTUP			WSAStartup;
  LPFN_WSACLEANUP			WSACleanup;
  LPFN_WSAGETLASTERROR			WSAGetLastError;
  LPFN_WSAWAITFORMULTIPLEEVENTS         WSAWaitForMultipleEvents;
  LPFN_WSACREATEEVENT			WSACreateEvent;
  LPFN_WSACLOSEEVENT			WSACloseEvent;
  LPFN_WSARESETEVENT			WSAResetEvent;
  LPFN_WSAEVENTSELECT			WSAEventSelect;
  LPFN_WSAENUMNETWORKEVENTS		WSAEnumNetworkEvents;
  LPFN_WSASEND				WSASend;

  LPFN_ACCEPT				accept;
  LPFN_BIND				bind;
  LPFN_CLOSESOCKET			closesocket;
  LPFN_CONNECT				connect;
  LPFN_IOCTLSOCKET			ioctlsocket;
  LPFN_GETSOCKOPT			getsockopt;
  LPFN_HTONL				htonl;
  LPFN_HTONS				htons;
  LPFN_INET_ADDR			inet_addr;
  LPFN_INET_NTOA			inet_ntoa;
  LPFN_LISTEN				listen;
  LPFN_NTOHS				ntohs;
  LPFN_NTOHL				ntohl;
  LPFN_RECV				recv;
  LPFN_SEND				send;
  LPFN_RECVFROM				recvfrom;
  LPFN_SENDTO				sendto;
  LPFN_SETSOCKOPT			setsockopt;
  LPFN_SHUTDOWN				shutdown;
  LPFN_SOCKET				socket;
  LPFN_GETHOSTBYNAME			gethostbyname;
  LPFN_GETHOSTBYADDR			gethostbyaddr;
  LPFN_GETHOSTNAME			gethostname;
  LPFN_GETSERVBYNAME			getservbyname;
  LPFN_GETSOCKNAME			getsockname;
  LPFN_GETPEERNAME			getpeername;
} Ws2Funcs;

#define ERRNO_BLOCK             WSAEWOULDBLOCK

#define EWOULDBLOCK             WSAEWOULDBLOCK
#define EINPROGRESS             WSAEINPROGRESS
#define EALREADY                WSAEALREADY
#define ENOTSOCK                WSAENOTSOCK
#define EDESTADDRREQ            WSAEDESTADDRREQ
#define EMSGSIZE                WSAEMSGSIZE
#define EPROTOTYPE              WSAEPROTOTYPE
#define ENOPROTOOPT             WSAENOPROTOOPT
#define EPROTONOSUPPORT         WSAEPROTONOSUPPORT
#define ESOCKTNOSUPPORT         WSAESOCKTNOSUPPORT
#define EOPNOTSUPP              WSAEOPNOTSUPP
#define EPFNOSUPPORT            WSAEPFNOSUPPORT
#define EAFNOSUPPORT            WSAEAFNOSUPPORT
#define EADDRINUSE              WSAEADDRINUSE
#define EADDRNOTAVAIL           WSAEADDRNOTAVAIL
#define ENETDOWN                WSAENETDOWN
#define ENETUNREACH             WSAENETUNREACH
#define ENETRESET               WSAENETRESET
#define ECONNABORTED            WSAECONNABORTED
#define ECONNRESET              WSAECONNRESET
#define ENOBUFS                 WSAENOBUFS
#define EISCONN                 WSAEISCONN
#define ENOTCONN                WSAENOTCONN
#define ESHUTDOWN               WSAESHUTDOWN
#define ETOOMANYREFS            WSAETOOMANYREFS
#define ETIMEDOUT               WSAETIMEDOUT
#define ECONNREFUSED            WSAECONNREFUSED
#define ELOOP                   WSAELOOP
#define ENAMETOOLONG            WSAENAMETOOLONG
#define EHOSTDOWN               WSAEHOSTDOWN
#define EHOSTUNREACH            WSAEHOSTUNREACH
#define ENOTEMPTY               WSAENOTEMPTY
#define EPROCLIM                WSAEPROCLIM
#define EUSERS                  WSAEUSERS
#define EDQUOT                  WSAEDQUOT
#define ESTALE                  WSAESTALE
#define EREMOTE                 WSAEREMOTE

#define INVALID_EVENT           WSA_INVALID_EVENT

typedef WSABUF sock_iovec;
#define iov_base buf
#define iov_len  len

static Ws2Funcs winSock;

#define sock_open(af, type, proto) (*winSock.socket)(af, type, proto)
#define sock_close(s)              (*winSock.closesocket)(s)
#define sock_shutdown(s,how)       (*winSock.shutdown)(s,how)
#define sock_accept(s, addr, len)  (*winSock.accept)(s, addr, len)
#define sock_connect(s, addr, len) (*winSock.connect)(s, addr, len)
#define sock_listen(s, b)          (*winSock.listen)(s, b)
#define sock_bind(s, addr, len)    (*winSock.bind)(s, addr, len)
#define sock_getopt(s,t,n,v,l)     (*winSock.getsockopt)(s,t,n,v,l)
#define sock_setopt(s,t,n,v,l)     (*winSock.setsockopt)(s,t,n,v,l)
#define sock_name(s, addr, len)    (*winSock.getsockname)(s, addr, len)
#define sock_peer(s, addr, len)    (*winSock.getpeername)(s, addr, len)
#define sock_ntohs(x)              (*winSock.ntohs)(x)
#define sock_ntohl(x)              (*winSock.ntohl)(x)
#define sock_htons(x)              (*winSock.htons)(x)
#define sock_htonl(x)              (*winSock.htonl)(x)
#define sock_send(s,buf,len,flag)  (*winSock.send)(s,buf,len,flag)
#define sock_sendv(s, vec, size, np, flag) \
                (*winSock.WSASend)(s,vec,size,np,flag,NULL,NULL)
#define sock_recv(s,buf,len,flag)  (*winSock.recv)(s,buf,len,flag)

#define sock_recvfrom(s,buf,blen,flag,addr,alen) \
                (*winSock.recvfrom)(s,buf,blen,flag,addr,alen)
#define sock_sendto(s,buf,blen,flag,addr,alen) \
                (*winSock.sendto)(s,buf,blen,flag,addr,alen)
#define sock_hostname(buf, len)    (*winSock.gethostname)(buf, len)

#define sock_errno()               (*winSock.WSAGetLastError)()
#define sock_create_event(d)       (*winSock.WSACreateEvent)()
#define sock_close_event(e)        (*winSock.WSACloseEvent)(e)

#define sock_select(d, flags, onoff) do { \
        (d)->event_mask = (onoff) ? \
                 ((d)->event_mask | (flags)) : \
                 ((d)->event_mask & ~(flags)); \
        (*winSock.WSAEventSelect)((d)->s, (d)->event, (d)->event_mask); \
  } while(0)

#define SET_BLOCKING(s)           (*winSock.ioctlsocket)(s, FIONBIO, &zero_value)
#define SET_NONBLOCKING(s)        (*winSock.ioctlsocket)(s, FIONBIO, &one_value)

static unsigned long zero_value = 0;
static unsigned long one_value = 1;

#else

#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef DEF_INADDR_LOOPBACK_IN_RPC_TYPES_H
#include <rpc/types.h>
#endif
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/uio.h>   /* for writev */
#include <netdb.h>

#ifdef USE_FIONBIO
#include <sys/ioctl.h>
static const int zero_value = 0, one_value = 1;
#define SET_BLOCKING(fd)        ioctl((fd), FIONBIO, &zero_value)
#define SET_NONBLOCKING(fd)     ioctl((fd), FIONBIO, &one_value)
#define ERRNO_BLOCK EWOULDBLOCK
#else

#include <fcntl.h>
#ifdef NB_O_NDELAY              /* Nothing needs this? */
#   define NB_FLAG O_NDELAY
#   ifndef ERRNO_BLOCK          /* allow override (e.g. EAGAIN) via Makefile */
#      define ERRNO_BLOCK EWOULDBLOCK
#   endif
#else  /* !NB_O_NDELAY */       /* The True Way - POSIX!:-) */
#   define NB_FLAG O_NONBLOCK
#   define ERRNO_BLOCK EAGAIN
#endif /* !NB_O_NDELAY */
#define SET_BLOCKING(fd)        fcntl((fd), F_SETFL, \
                                      fcntl((fd), F_GETFL, 0) & ~NB_FLAG)
#define SET_NONBLOCKING(fd)     fcntl((fd), F_SETFL, \
                                      fcntl((fd), F_GETFL, 0) | NB_FLAG)
#endif

#define INVALID_SOCKET -1
#define INVALID_EVENT  -1
#define SOCKET_ERROR   -1
#define SOCKET int
#define HANDLE int
#define FD_READ    DO_READ
#define FD_WRITE   DO_WRITE
#define FD_CLOSE   0
#define FD_CONNECT DO_WRITE
#define FD_ACCEPT  DO_READ

#ifndef SD_SEND
#define SD_RECV 0
#define SD_SEND 1
#define SD_BOTH 2
#endif


typedef struct iovec sock_iovec;

extern int close();
extern int gethostname();

#define sock_open(af, type, proto)  socket(af, type, proto)
#define sock_close(s)               close(s)
#define sock_shutdown(s,how)        shutdown(s,how)
#define sock_accept(s, addr, len)   accept(s, addr, len)
#define sock_connect(s, addr, len)  connect(s, addr, len)
#define sock_listen(s, b)           listen(s, b)
#define sock_bind(s, addr, len)     bind(s, addr, len)
#define sock_getopt(s,t,n,v,l)      getsockopt(s,t,n,v,l)
#define sock_setopt(s,t,n,v,l)      setsockopt(s,t,n,v,l)
#define sock_name(s, addr, len)     getsockname(s, addr, len)
#define sock_peer(s, addr, len)     getpeername(s, addr, len)
#define sock_ntohs(x)               ntohs(x)
#define sock_ntohl(x)               ntohl(x)
#define sock_htons(x)               htons(x)
#define sock_htonl(x)               htonl(x)
#define sock_send(s,buf,len,flag)   send(s,buf,len,flag)
#define sock_sendv(s, vec, size, np, flag) \
		(*(np) = writev(s, vec, size))
#define sock_recv(s,buf,len,flag)   recv(s,buf,len,flag)
#define sock_recvfrom(s,buf,blen,flag,addr,alen) \
                recvfrom(s,buf,blen,flag,addr,alen)
#define sock_sendto(s,buf,blen,flag,addr,alen) \
                sendto(s,buf,blen,flag,addr,alen)
#define sock_hostname(buf, len)    gethostname(buf, len)

#define sock_errno()                errno
#define sock_create_event(d)        (d)->s   /* return file descriptor */
#define sock_close_event(e)                  /* do nothing */

#define sock_select(d, flags, onoff) do { \
        (d)->event_mask = (onoff) ? \
                 ((d)->event_mask | (flags)) : \
                 ((d)->event_mask & ~(flags)); \
        driver_select((d)->port, (d)->event, (flags), (onoff)); \
   } while(0)

#endif


#ifndef USE_THREADS
#error "etk_drv must be compiled with threads"
#endif


#if defined(POSIX_THREADS)
#include <pthread.h>

typedef pthread_t       THREAD_T;
typedef pthread_mutex_t LOCK_T;

#elif defined(SOLARIS_THREADS)
#include <thread.h>

typedef thread_t        THREAD_T;
typedef mutex_t         LOCK_T;

#elif defined(__WIN32__)

typedef HANDLE           THREAD_T;
typedef CRITICAL_SECTION LOCK_T;

#else
#error "bad thread configuration"
#endif

#define ALIGN_SZ	sizeof(double)
#define MESG_HDR_SZ	(sizeof(erltk_mesg_t)-ALIGN_SZ)
#define MESG_BUF_SZ     256	/* must be dividable with ALIGN_SZ!!! */
#define MESG_FIXED_SZ	(MESG_HDR_SZ + MESG_BUF_SZ)
#define MESG_NO_FIXED	4	/* number of fixed messages */

typedef unsigned char u_int8;

typedef struct erltk_mesg {
    struct erltk_mesg* next; /* next in free list */
    int dynamic;			 /* fixed message or not */
    int len;
    u_int8 data[ALIGN_SZ];	 /* use sizeof(double) for alignment */
} erltk_mesg_t;

typedef struct erltk_desc {
    SOCKET s;          /* Erlang side socket or INVALID_SOCKET if not open */
    SOCKET s_tk;       /* Tk side socket or INVALID if not open */

    HANDLE event;      /* Event handle (same as s in unix) */
    long  event_mask;  /* current FD events */
    long  port;        /* the port identifier */
} erltk_desc_t;

/* allocate a pool of fix sized messages (each message must be of size=N bytes
This should fix some memory allocation overhead problems.	
*/

static int  erltk_finish();
static int  erltk_init();
static long erltk_start();
static int  erltk_stop();
static int  erltk_command();
static int  erltk_input();
static int  erltk_output();
#ifdef __WIN32__
static int  erltk_event();
#endif

DriverEntry tk_driver_entry =
{
    erltk_init,
    erltk_start,
    erltk_stop,
    erltk_command,
#ifdef __WIN32__
    erltk_event,
#else
    erltk_input,
#endif
    erltk_output,
    "etk_drv",
    NULL,		/* handle */
    NULL,		/* ctrl */
    NULL,		/* timeout */
    NULL		/* outputv */
};

static erltk_desc_t* erltk_desc;
static char*         erltk_port_arg;  /* must be freed */
static THREAD_T      erltk_tid;	     /* The thread */
/* Lock to avoid erlang/otp to unload the driver while
** tk is still terminating.
** The lock is taken in ErlTk_Init when tk starts execute 
** and is released in ErlTK_Exit.
*/
static LOCK_T       erltk_lock;

/* #define DEBUG  */
/* #define TRACE  */

#ifdef DEBUG
#define DPRINT(fmt, arg) printf(fmt, arg)
#else
#define DPRINT(fmt, arg)
#endif

#ifdef TRACE
#define TPRINT(fmt, a1)     printf(fmt, a1)
#define TPRINT2(fmt, a1,a2) printf(fmt, a1,a2)
#else
#define TPRINT(fmt, a1)
#define TPRINT2(fmt, a1, a2)
#endif

#define ERLTK_MESG_OK      0
#define ERLTK_MESG_CLOSED  1
#define ERLTK_MESG_ERROR   2

static int erl_closed = 0;

#ifndef __WIN32__
extern int matherr();
int *tclDummyMathPtr = (int *) matherr;
#endif

EXTERN int Tcl_LinkVar();
int (*tclDummyLinkVarPtr)() = Tcl_LinkVar;

/* The main window for the application.  If
 * NULL then the application no longer
 * exists. */
static Tk_Window mainWindow = NULL; 

static Tcl_Interp *interp = NULL; /* Interpreter for this application. */
static int synchronize;
static int withdrawn;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;
static char *colormap;
static char *visual;
static int rest = 0;

static Tk_ArgvInfo argTable[] = {
    {"-colormap", TK_ARGV_STRING, (char *) NULL, (char *) &colormap,
        "Colormap for main window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
        "Display to use"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
        "Initial geometry for window"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
        "Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
        "Use synchronous mode for display server"},
    {"-visual", TK_ARGV_STRING, (char *) NULL, (char *) &visual,
        "Visual for main window"},
    /* Added for Erltk */
    {"-withdrawn", TK_ARGV_CONSTANT, (char*) 1, (char *) &withdrawn,
	 "Start erlkt main window withdrawn" },
    {"--", TK_ARGV_REST, (char *) 1, (char *) &rest,
	 "Pass all remaining arguments through to script"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	 (char *) NULL}
};



/* --------------------------------------------------------------------------

   TK DRIVER THREAD STUFF

 --------------------------------------------------------------------------*/

static int ThreadsInit()
{
    /* Dummy (but may be used for other thread packages) */
	return 0;
}

/* Block some signals to this thread */
static int ThreadBlockSignals()
{
#if defined(POSIX_THREADS)
    sigset_t new;

    sigemptyset(&new);
    sigaddset(&new, SIGINT);   /* block interrupt */
    sigaddset(&new, SIGCHLD);  /* block pipe signals */
    return pthread_sigmask(SIG_BLOCK, &new, NULL);
#elif defined(SOLARIS_THREADS)
    sigset_t new;

    sigemptyset(&new);
    sigaddset(&new, SIGINT);   /* block interrupt */
    sigaddset(&new, SIGCHLD);  /* block pipe signals */
    return thr_sigsetmask(SIG_BLOCK, &new, NULL);
#else
    return 0;
#endif
}

/* create a thread */
static int ThreadSpawn(func, param, tid)
void* (*func)(void*); void* param; THREAD_T* tid;
{
#if defined(POSIX_THREADS)
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    return pthread_create(tid, &attr, func, param);
#elif defined(SOLARIS_THREADS)
    return thr_create(NULL, 0, func, param,
		      THR_BOUND | THR_NEW_LWP | THR_DETACHED, tid);
#elif defined(__WIN32__)
    HANDLE h;
    DWORD ID;

    h = (HANDLE) _beginthreadex(NULL, 0, (LPTHREAD_START_ROUTINE) func, 
								(LPVOID)param, 0, &ID);
    if (h == INVALID_HANDLE_VALUE)
		return -1;
    CloseHandle(h);
	*tid = h;
	return 0;
#endif
}

/* wait for thread with id to terminate */
static int ThreadJoin(tid, result)
THREAD_T tid; void** result;
{
#if defined(POSIX_THREADS)
    return pthread_join(tid, result);
#elif defined(SOLARIS_THREADS)
    return thr_join(tid, NULL, result);
#elif defined(__WIN32__)
    WaitForSingleObject(tid, INFINITE); /* FIX ERRORS */
    *result = 0;
    CloseHandle(tid);
    return 0;
#endif
}


/* Brutal stop */
static int ThreadKill(tid)
THREAD_T tid;
{
#if defined(POSIX_THREADS)
    return pthread_kill(tid, SIGINT);
#elif defined(SOLARIS_THREADS)
    return thr_kill(tid, SIGINT);
#elif defined(__WIN32__)
    return -1;
#endif
}


/* Terminate current thread */
static void ThreadExit(result)
void* result;
{
#if defined(POSIX_THREADS)
    pthread_exit(result);
#elif defined(SOLARIS_THREADS)
    thr_exit(result);
#elif defined(__WIN32__)
    _endthreadex((unsigned) result);
#endif
}

static int LockInit(lockp)
LOCK_T* lockp;
{
#if defined(POSIX_THREADS)
    return pthread_mutex_init(lockp, NULL);
#elif defined(SOLARIS_THREADS)
    return mutex_init(lockp, USYNC_THREAD, NULL);
#elif defined(__WIN32__)
    InitializeCriticalSection(lockp);
	return 0;
#endif
}

static int LockDestroy(lockp)
LOCK_T* lockp;
{
#if defined(POSIX_THREADS)
    pthread_mutex_destroy(lockp);
#elif defined(SOLARIS_THREADS)
    mutex_destroy(lockp);
#elif defined(__WIN32__)
    DeleteCriticalSection(lockp);
	return 0;
#endif
}

static int Lock(lockp)
LOCK_T* lockp;
{
#if defined(POSIX_THREADS)
    pthread_mutex_lock(lockp);
#elif defined(SOLARIS_THREADS)
    mutex_lock(lockp);
#elif defined(__WIN32__)
    EnterCriticalSection(lockp);
	return 0;
#endif
}

static int Unlock(lockp)
LOCK_T* lockp;
{
#if defined(POSIX_THREADS)
    pthread_mutex_lock(lockp);
#elif defined(SOLARIS_THREADS)
    mutex_unlock(lockp);
#elif defined(__WIN32__)
    LeaveCriticalSection(lockp);
	return 0;
#endif
}

/* --------------------------------------------------------------------------

   MESSAGE STUFF

 --------------------------------------------------------------------------*/

static LOCK_T message_lock;
static erltk_mesg_t* message_pool;
static char mesg_buffer[MESG_NO_FIXED*MESG_FIXED_SZ];

static void init_message_pool()
{
    char* ptr = mesg_buffer;
    erltk_mesg_t* mesg;
    int i;

    LockInit(&message_lock);
    message_pool = NULL;
    for (i = 0; i < MESG_NO_FIXED; i++) {
	mesg = (erltk_mesg_t*) ptr;
	ptr += MESG_FIXED_SZ;
	mesg->next = message_pool;
	message_pool = mesg;
    }
}

static erltk_mesg_t* alloc_message(len)
int len;
{
    erltk_mesg_t* mesg = NULL;

    if (len <= MESG_BUF_SZ) {
	Lock(&message_lock);
	if (message_pool != NULL) {
	    mesg = message_pool;
	    mesg->dynamic = 0;
	    message_pool = mesg->next;
	}
	Unlock(&message_lock);
    }
    if (mesg == NULL) {
	mesg = (erltk_mesg_t*) malloc(MESG_HDR_SZ+len);
	if (mesg == NULL)
	    return NULL;
	mesg->dynamic = 1;
    }
    mesg->len = len;
    return mesg;
}

static void free_message(mesg)
erltk_mesg_t* mesg;
{
    if (mesg->dynamic)
	free(mesg);
    else {
	Lock(&message_lock);
	mesg->next = message_pool;
	message_pool = mesg;
	Unlock(&message_lock);
    }
}

static erltk_mesg_t* realloc_message(mesg, len)
erltk_mesg_t* mesg; int len;
{
    if (len > mesg->len) {
	if (mesg->dynamic || (len > MESG_BUF_SZ)) {
	    free_message(mesg);
	    return alloc_message(len);
	}
    }
    mesg->len = len;
    return mesg;
}

/* --------------------------------------------------------------------------

   TK DRIVER STUFF

 --------------------------------------------------------------------------*/

static int ErlTk_SendMsg(s, mesg)
SOCKET s; erltk_mesg_t* mesg;
{
    int n;
	
    n = sock_send(s, (char*) &mesg, sizeof(erltk_mesg_t*), 0);
    if (n != sizeof(erltk_mesg_t*)) {
	free_message(mesg);
	return -1;
    }
    return 0;
}

static erltk_mesg_t* ErlTk_RecvMsg(s, err)
SOCKET s; int* err;
{
    erltk_mesg_t* mesg;
    int n;

    if ((n = sock_recv(s,(char*)&mesg,sizeof(erltk_mesg_t*),0)) == 0) {
	*err = ERLTK_MESG_CLOSED;
	return NULL;
    }
    else if (n != sizeof(erltk_mesg_t*)) {
	*err = ERLTK_MESG_ERROR;
	return NULL;
    }
    *err = ERLTK_MESG_OK;
    return mesg;
}

/* Create socket pair i.e something like a bidirectional pipe */
static int sockpair(SOCKET pair[2])
{
    struct sockaddr_in a;
    SOCKET s0 = INVALID_SOCKET;
    SOCKET s1 = INVALID_SOCKET;
    SOCKET s2 = INVALID_SOCKET;
    int len = sizeof(a);
    int on = 1;

    if ((s0 = sock_open(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	goto error;

    memset((char*) &a, 0, sizeof(a));
    a.sin_family = AF_INET;
    a.sin_addr.s_addr = sock_htonl(INADDR_LOOPBACK);
    a.sin_port = sock_htons(0);

    if (sock_bind(s0,(struct sockaddr*) &a,sizeof(a)) == SOCKET_ERROR)
	goto error;

    if (sock_name(s0, (struct sockaddr*) &a, &len) == SOCKET_ERROR)
	goto error;

    if (sock_listen(s0, 1) == SOCKET_ERROR)
	goto error;

    if ((s1 = sock_open(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	goto error;

    if (sock_connect(s1, (struct sockaddr*) &a, len) == SOCKET_ERROR)
	goto error;

    if ((s2 = sock_accept(s0, (struct sockaddr*) &a, &len)) == INVALID_SOCKET)
	goto error;
    sock_close(s0);
    /* The TCP_NODELAY option are needed */ 
    sock_setopt(s1, IPPROTO_TCP, TCP_NODELAY, (char*) &on, sizeof(int));
    sock_setopt(s2, IPPROTO_TCP, TCP_NODELAY, (char*) &on, sizeof(int));
    SET_NONBLOCKING(s1);
    SET_NONBLOCKING(s2);
    pair[0] = s1;
    pair[1] = s2;
    return 0;

error:
    if (s0 != INVALID_SOCKET)
	sock_close(s0);
    if (s1 != INVALID_SOCKET)
	sock_close(s1);
    if (s2 != INVALID_SOCKET)
	sock_close(s2);
    return -1;
}


/*
** Must map exit => thr_exit
*/
static void ErlTk_Exit(code, msg)
int code; char* msg;
{
    erltk_mesg_t* mesg;
    int reason;
    DPRINT("ErlTk_Exit: %s\n\r", msg);

    if (erltk_port_arg != NULL) {
	free(erltk_port_arg);
	erltk_port_arg = NULL;
    }
    sock_shutdown(erltk_desc->s_tk,SD_SEND);
    while((mesg = ErlTk_RecvMsg(erltk_desc->s_tk,&reason)) != NULL)
	free_message(mesg);
    sock_close(erltk_desc->s_tk);
    erl_closed = 1;
    if (mainWindow != NULL)
	Tk_DestroyWindow(mainWindow);
    if (interp != NULL)
	Tcl_DeleteInterp(interp);
    Unlock(&erltk_lock);   /* locked while executing */
    /* A tiny hole */
    ThreadExit(code);
}

/* erlang function are descibed by fun:id */
/* the function it self is stored in an ets data base */

static int ErlTk_Execute(cmd, data, interp, argc, argv)
int cmd; ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
#ifdef TRACE
    int i;

    TPRINT("ErlTk_Execute:", 0);
    for (i = 0; i < argc; i++)
	TPRINT(" %s", argv[i]);
    TPRINT("\n\r", 0);
#endif
    switch(cmd) {
    case ERL_TK_BIND:
	return Tk_BindCmd(data, interp, argc, argv);

    case ERL_TK_DESTROY:
	return Tk_DestroyCmd(data, interp, argc, argv);

    case ERL_TK_LOWER:
	return Tk_LowerCmd(data, interp, argc, argv);
	
    case ERL_TK_RAISE:
	return Tk_RaiseCmd(data, interp, argc, argv);
	
    case ERL_TK_BELL:
	return Tk_BellCmd(data, interp, argc, argv);
	
    case ERL_TK_BUTTON:
	return Tk_ButtonCmd(data, interp, argc, argv);
	
    case ERL_TK_CHECKBUTTON:
	return Tk_CheckbuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_RADIOBUTTON:
	return Tk_RadiobuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_LABEL:
	return Tk_LabelCmd(data, interp, argc, argv);
	
    case ERL_TK_UPDATE:
	return Tk_UpdateCmd(data, interp, argc, argv);
	
    case ERL_TK_WINFO:
	return Tk_WinfoCmd(data, interp, argc, argv);
	
    case ERL_TK_WM:
	return Tk_WmCmd(data, interp, argc, argv);
	
    case ERL_TK_BINDTAGS:
	return Tk_BindtagsCmd(data, interp, argc, argv);
	
    case ERL_TK_CANVAS:
	return Tk_CanvasCmd(data, interp, argc, argv);
	
    case ERL_TK_CLIPBOARD:
	return Tk_ClipboardCmd(data, interp, argc, argv);
	
    case ERL_TK_ENTRY:
	return Tk_EntryCmd(data, interp, argc, argv);
	
    case ERL_TK_FRAME:
	return Tk_FrameCmd(data, interp, argc, argv);

    case ERL_TK_TOPLEVEL:
	return Tk_ToplevelCmd(data, interp, argc, argv);
	
    case ERL_TK_FOCUS:
	return Tk_FocusCmd(data, interp, argc, argv);
	
    case ERL_TK_GRAB:
	return Tk_GrabCmd(data, interp, argc, argv);
	
    case ERL_TK_IMAGE:
	return Tk_ImageCmd(data, interp, argc, argv);
	
    case ERL_TK_LISTBOX:
	return Tk_ListboxCmd(data, interp, argc, argv);
	
    case ERL_TK_MENU:
	return Tk_MenuCmd(data, interp, argc, argv);
	
    case ERL_TK_MENUBUTTON:
	return Tk_MenubuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_MESSAGE:
	return Tk_MessageCmd(data, interp, argc, argv);
	
    case ERL_TK_OPTION:
	return Tk_OptionCmd(data, interp, argc, argv);
	
    case ERL_TK_PACK:
	return Tk_PackCmd(data, interp, argc, argv);
	
    case ERL_TK_PLACE:
	return Tk_PlaceCmd(data, interp, argc, argv);
	
    case ERL_TK_SCALE:
	return Tk_ScaleCmd(data, interp, argc, argv);
	
    case ERL_TK_SCROLLBAR:
	return Tk_ScrollbarCmd(data, interp, argc, argv);
	
    case ERL_TK_SELECTION:
	return Tk_SelectionCmd(data, interp, argc, argv);
	
    case ERL_TK_TEXT:
	return Tk_TextCmd(data, interp, argc, argv);

    case ERL_TK_GRID:
	return Tk_GridCmd(data, interp, argc, argv);
	
    case ERL_TK_TK:
	return Tk_TkCmd(data, interp, argc, argv);
	
    case ERL_TK_CMD: {
	Tcl_CmdInfo info;

	if (Tcl_GetCommandInfo(interp, argv[0], &info))
	    return (*info.proc)(info.clientData, interp, argc, argv);
	else {
	    Tcl_SetResult(interp, "widget not found", TCL_STATIC);
	    return TCL_ERROR;
	}
    }

    case ERL_TK_GETVAR: {
	char *value = Tcl_GetVar(interp, argv[1], TCL_GLOBAL_ONLY);
	if (value == NULL)
	    value = "";
	Tcl_AppendResult(interp, value, (char*) NULL);
	return TCL_OK;
    }

    case ERL_TK_SETVAR:
	if (Tcl_SetVar(interp, argv[1], argv[2],
		       TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
	    return TCL_ERROR;
	}
	return TCL_OK;

    case ERL_TK_EVENT:
	return Tk_EventCmd(data, interp, argc, argv);

    default:
	ErlTk_Exit(1, "unknown erltk command");
	return 0;
    }
}

/*
** Send argv info to erlang as:
**
**  32-bit-length  data
*/

static int SendToErlang(cmd, clientData, interp, argc, argv)
int cmd; ClientData clientData; Tcl_Interp* interp; int argc; char** argv;
{
    int i;
    int len;
    u_int8* ptr;
    erltk_mesg_t* mesg;

    if (erl_closed)
	return TCL_OK;

    /* calc total length */
    len = 0;
    for (i = 1; i < argc; i++)
	len += strlen(argv[i]) + 4;

    mesg = alloc_message(len+4);
    ptr = mesg->data;

    PUT_INT16(ptr, 0); 
    ptr += 2;
    PUT_INT16(ptr, cmd);
    ptr += 2;

    for (i = 1; i < argc; i++) {
	int n = strlen(argv[i]);
	PUT_INT32(ptr, n); ptr += 4;
	memcpy(ptr, argv[i], n);
	ptr += n;
    }
    ErlTk_SendMsg(erltk_desc->s_tk, mesg);
    return TCL_OK;
}

/*
** Callback for tkerror & bgerror
*/
static int tkError(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_TKERROR, data, interp, argc, argv);
}

/*
** Callback for tkscreenchanged
*/
static int tkScreenChanged(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_TKSCREEN, data, interp, argc, argv);
}

/*
** Callback for event functions
** args:  event Id argv[2] argv[3] ... 
** write  Id, argv[2], argv[3] ...
*/

static int ErlangEvent(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_EVENT, data, interp, argc, argv);
}

/*
** Callback for invoke functions
** args:  invoke Id argv[2] argv[3] ...
** write  Id, argv[2], argv[3] ...
*/

static int ErlangInvoke(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_INVOKE, data, interp, argc, argv);
}


/*
**  INPUT:
**  +-----------------+
**  | Sequence number |   2 bytes
**  +-----------------+
**  | Command         |   2 bytes
**  +-----------------+
**  | Argc = n        |   4 bytes
**  +-----------------+
**  | OP | Argv[0]    |   4 bytes
**  +-----------------+
**  | OP | Argv[1]    |   4 bytes
**  +-----------------+
**  |       ...       |
**  +-----------------+
**  | OP | Argv[n-1]  |   4 bytes
**  +-----------------+
**  |  0   0   0   0  |
**  +-----------------+
**  | Len(4) Data     |
**  | Len(4) Data     |
**  | ...             |
**  |                 |
**  +-----------------+
**
**  OUTPUT:
**  +-----------------+
**  | Sequence number |   2 bytes
**  +-----------------+
**  | Reply code      |   2 bytes
**  +-----------------+
**  | Reply string    |
**  |     ...         |
**  +-----------------+
**
*/

static void ErlangInputProc(data, mask)
ClientData data; int mask;
{
    int len;
    int argc;
    int voffs;
    int cmd;
    int seq;
    int code;
    int reason;
    int i;
    u_int8* p;
    u_int8* ptr;
    char** argv;
    erltk_mesg_t* mesg;

    if ((mesg = ErlTk_RecvMsg(erltk_desc->s_tk,&reason)) == NULL) {
	if (reason == ERLTK_MESG_CLOSED)
	    ErlTk_Exit(0, "eof");
	else
	    ErlTk_Exit(1, "input error");
    }

    TPRINT2("ErlangInputProc: message(%s), len = %d\n\r", 
		mesg->dynamic?"dynaic":"fixed", mesg->len);

    len = mesg->len;
    p = mesg->data;
    seq = GET_INT16(p); p += 2;
    cmd = GET_INT16(p); p += 2;
    argc = GET_INT32(p); p += 4;
    argv = (char**) (p);
    ptr  = p + 4*(argc+1);   /* data area pointer */

    voffs = argc*4;
    for (i = 0; i < argc; i++) {
	int offs;

	p = (u_int8*) (argv + i);
	switch (*p++) {
	case TK_OP_OFFSET:
	    len = GET_INT32(ptr);  /* get data length */
	    ptr += 4;              /* skip length field */
	    argv[i] = ptr;
	    ptr += len;            /* prepare for next data */
	    break;
	case TK_OP_OPTION:
	    offs = GET_INT24(p);
	    if (offs >= OPT_MAX)
		ErlTk_Exit(1, "bad option offset");
	    argv[i] = tkopt[offs];
	    break;
	case TK_OP_STRING:
	    offs = GET_INT24(p);
	    if (offs >= STR_MAX)
		ErlTk_Exit(1, "bad string offset");
	    argv[i] = tkstr[offs];
	    break;
	default:
	    ErlTk_Exit(1, "bad option type");
	}
    }

    switch(ErlTk_Execute(cmd,(ClientData)mainWindow,interp,argc,argv)) {
    case TCL_BREAK:
	free_message(mesg);
	Tcl_ResetResult(interp);
	return;

    case TCL_OK:
	code = TK_ERL_OK;
	break;

    case TCL_ERROR:
	code = TK_ERL_ERROR;
	break;
    }

    len = strlen(interp->result);

    TPRINT("Result: %s\n\r", interp->result);

    mesg = realloc_message(mesg, len+4);

    PUT_INT16(mesg->data, seq);
    PUT_INT16(mesg->data+2, code);
    memcpy(mesg->data+4, interp->result, len);
    Tcl_ResetResult(interp);

    ErlTk_SendMsg(erltk_desc->s_tk, mesg);
}

#ifdef __WIN32__

#ifndef TCL_ACTIVE
#define TCL_ACTIVE (1<<4)
#endif

extern void TkWinXInit(HINSTANCE hInstance);

static void DisplaySetupProc(ClientData clientData, int flags)
{
    if ((flags & TCL_WINDOW_EVENTS) && (Tk_GetNumMainWindows()>0)) {
	Tcl_WatchFile((Tcl_File) NULL, TCL_ACTIVE);
    }
}

static void DisplayCheckProc(ClientData clientData, int flags)
{
    return;
}

static void TkCreateWinEventSource()
{
    Tcl_CreateEventSource(DisplaySetupProc, DisplayCheckProc,
			  (ClientData) NULL);
    TkWinXInit((HINSTANCE)tk_driver_entry.handle);
}

#endif

/*
** Port main loop
*/

static void ErlTk_Main(argc, argv)
int argc; char** argv;
{
    /* Initialize */
    int code;
    char* av[20];
    int   ac;
    char* class;

    erl_closed = 0;
    interp = Tcl_CreateInterp();

    synchronize = 0;
    name = display = geometry = colormap = visual = NULL; 

    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 
		     TK_ARGV_NO_DEFAULTS) != TCL_OK) {
	ErlTk_Exit(1, interp->result);
    }

    if (name == NULL) {
	char* p = strrchr(argv[0], '/');
	if (p == NULL)
	    name = argv[0];
	else
	    name = p+1;
    }

    class = (char *) ckalloc((unsigned) (strlen(name) + 1));
    strcpy(class, name);
    class[0] = toupper((unsigned char) class[0]);

    av[0] = "toplevel";
    av[1] = ".";
    av[2] = "-class";
    av[3] = class;
    ac = 4;
 
    if (display != NULL) {
	av[ac] = "-screen";
	av[ac+1] = display;
	ac += 2;
    }
    if (colormap != NULL) {
	av[ac] = "-colormap";
	av[ac+1] = colormap;
	ac += 2;
    }
    if (visual != NULL) {
		av[ac] = "-visual";
		av[ac+1] = visual;
		ac += 2;
    }
    av[ac] = NULL;
    code = TkCreateFrame((ClientData) NULL, interp, ac, av, 1, name);

    ckfree(class);
    if (code != TCL_OK)
	ErlTk_Exit(1, interp->result);
    mainWindow = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "event", 
		      (Tcl_CmdProc*) ErlangEvent, 
		      (ClientData) 0, (Tcl_CmdDeleteProc*) NULL);

    Tcl_CreateCommand(interp, "invoke", 
		      (Tcl_CmdProc*) ErlangInvoke, 
		      (ClientData) 0, (Tcl_CmdDeleteProc*) NULL);

    Tcl_CreateCommand(interp, "bgerror", 
		      (Tcl_CmdProc*) tkError, 
		      (ClientData) 0, (Tcl_CmdDeleteProc*) NULL);

    Tcl_CreateCommand(interp, "tkScreenChanged", 
		      (Tcl_CmdProc*) tkScreenChanged, 
		      (ClientData) 0, (Tcl_CmdDeleteProc*) NULL);
    Tcl_ResetResult(interp);

#ifndef __WIN32__
    if (synchronize) {
	XSynchronize(Tk_Display(mainWindow), True);
    }    
#endif
    if (geometry != NULL) {
	av[0] = "wm";
	av[1] = "geometry";
	av[2] = ".";
	av[3] = geometry;
	av[4] = NULL;

	if (Tk_WmCmd((ClientData) mainWindow, interp, 4, av) != TCL_OK)
	    ErlTk_Exit(1, interp->result);
    }

    if (withdrawn) {
	av[0] = "wm";
	av[1] = "withdraw";
	av[2] = ".";
	av[3] = NULL;
	if (Tk_WmCmd((ClientData) mainWindow, interp, 3, av) != TCL_OK) {
	    ErlTk_Exit(1, interp->result);
	}
    }
#ifdef __WIN32__
    TkCreateWinEventSource();
    Tcl_CreateFileHandler(Tcl_GetFile((ClientData)erltk_desc->s_tk,
				      TCL_WIN_SOCKET),
			  TK_READABLE, (Tcl_FileProc*) ErlangInputProc,
			  (ClientData) 0);
#else
    TkCreateXEventSource();
    Tcl_CreateFileHandler(Tcl_GetFile((ClientData)erltk_desc->s_tk,
				      TCL_UNIX_FD),
			  TK_READABLE, (Tcl_FileProc*) ErlangInputProc,
			  (ClientData) 0);
#endif

    while (Tk_GetNumMainWindows() > 0) {
        Tcl_DoOneEvent(0);
    }
    ErlTk_Exit(0, "done");
}


#define ERLTK_MAX_ARGV 10

#define IS_BLANK(c) ( ((c) <= ' ') && \
                      (((c) == ' ') || ((c) == '\t') || ((c) == '\n')))

/* Create argument vectors etc, call ErlTk_Main */
static void* ErlTk_Init(arg)
void* arg;
{
    char* argv[ERLTK_MAX_ARGV+1];
    char* ptr;
    int argc;
    int c;

    ThreadBlockSignals();
    Lock(&erltk_lock);   /* lock while executing */

    argv[0] = "erltk";
    argc = 1;
    ptr = (char*) arg;

    while(argc < ERLTK_MAX_ARGV) {
	c = *ptr;
	while(IS_BLANK(c))
	    c = *++ptr;  /* skip leading blanks */
	if (c == 0)
	    break;
	argv[argc] = ptr;   /* set argument */
	c = *++ptr;
	while((c != 0) && !IS_BLANK(c))
	    c = *++ptr;
	argc++;
	if (c == 0)
	    break;
	else
	    *ptr++ = 0;
    }
    argv[argc] = 0;
    ErlTk_Main(argc, argv);
    return 0;
}

#ifdef __WIN32__
/* Called from tclWinSocket.c !!! */

HINSTANCE ErlTk_GetInstance()
{
    return (HINSTANCE) tk_driver_entry.handle;
}
#endif


/* since tk is not reentrant (its compiled so) we may only have
   one copy of tk active !!!
*/

/* Loadable driver */ 
#ifdef LOADABLE
int DRIVER_INIT(etk_drv)(void* handle)
{
    tk_driver_entry.handle = handle;
    tk_driver_entry.driver_name = "etk_drv";
    tk_driver_entry.finish = erltk_finish;
    tk_driver_entry.init = erltk_init;
    tk_driver_entry.start = erltk_start;
    tk_driver_entry.stop = erltk_stop;
    tk_driver_entry.output = erltk_command;
#ifdef __WIN32__
    tk_driver_entry.ready_input = erltk_event;
#else
    tk_driver_entry.ready_input = erltk_input;
#endif
    tk_driver_entry.ready_output = erltk_output;
    tk_driver_entry.timeout = NULL;
    tk_driver_entry.outputv = NULL;
    tk_driver_entry.control = NULL;
    return (int) &tk_driver_entry;
}
#endif


#ifdef __WIN32__

static int lookup_functions(void)
{
    static char dll_name[] = "ws2_32";
    HMODULE module;

    if (winSock.WSAStartup != NULL)
	return TRUE;

    module = LoadLibrary(dll_name);
    
    if (module == NULL)
	return FALSE;

    winSock.WSAStartup = (LPFN_WSASTARTUP) GetProcAddress(module, "WSAStartup");
    winSock.WSACleanup = (LPFN_WSACLEANUP) GetProcAddress(module, "WSACleanup");
    winSock.WSAGetLastError = (LPFN_WSAGETLASTERROR) GetProcAddress(module, "WSAGetLastError");
    winSock.WSAWaitForMultipleEvents = (LPFN_WSAWAITFORMULTIPLEEVENTS) GetProcAddress(module, "WSAWaitForMultipleEvents");
    winSock.WSACreateEvent = (LPFN_WSACREATEEVENT) GetProcAddress(module, "WSACreateEvent");

    winSock.WSACloseEvent = (LPFN_WSACLOSEEVENT) GetProcAddress(module, "WSACloseEvent");
    winSock.WSAResetEvent = (LPFN_WSARESETEVENT) GetProcAddress(module, "WSAResetEvent");
    winSock.WSAEventSelect = (LPFN_WSAEVENTSELECT) GetProcAddress(module, "WSAEventSelect");
    winSock.WSAEnumNetworkEvents = (LPFN_WSAENUMNETWORKEVENTS) GetProcAddress(module, "WSAEnumNetworkEvents");
    winSock.WSASend = (LPFN_WSASEND) GetProcAddress(module, "WSASend");
    
    winSock.accept = (LPFN_ACCEPT) GetProcAddress(module, "accept");
    winSock.bind = (LPFN_BIND) GetProcAddress(module, "bind");
    winSock.closesocket = (LPFN_CLOSESOCKET) GetProcAddress(module, "closesocket");
    winSock.connect = (LPFN_CONNECT) GetProcAddress(module, "connect");
    winSock.ioctlsocket = (LPFN_IOCTLSOCKET) GetProcAddress(module, "ioctlsocket");

    winSock.getsockopt = (LPFN_GETSOCKOPT) GetProcAddress(module, "getsockopt");
    winSock.htonl = (LPFN_HTONL) GetProcAddress(module, "htonl");
    winSock.htons = (LPFN_HTONS) GetProcAddress(module, "htons");
    winSock.inet_addr = (LPFN_INET_ADDR) GetProcAddress(module, "inet_addr");
    winSock.inet_ntoa = (LPFN_INET_NTOA) GetProcAddress(module, "inet_ntoa");
    winSock.listen = (LPFN_LISTEN) GetProcAddress(module, "listen");
    winSock.ntohs = (LPFN_NTOHS) GetProcAddress(module, "ntohs");
    winSock.ntohl = (LPFN_NTOHL) GetProcAddress(module, "ntohl");
    winSock.recv = (LPFN_RECV) GetProcAddress(module, "recv");
    winSock.send = (LPFN_SEND) GetProcAddress(module, "send");
    winSock.recvfrom = (LPFN_RECVFROM) GetProcAddress(module, "recvfrom");
    winSock.sendto = (LPFN_SENDTO) GetProcAddress(module, "sendto");
    winSock.setsockopt = (LPFN_SETSOCKOPT) GetProcAddress(module, "setsockopt");
    winSock.shutdown = (LPFN_SHUTDOWN) GetProcAddress(module, "shutdown");
    winSock.socket = (LPFN_SOCKET) GetProcAddress(module, "socket");
    winSock.gethostbyaddr = (LPFN_GETHOSTBYADDR) GetProcAddress(module, "gethostbyaddr");
    winSock.gethostbyname = (LPFN_GETHOSTBYNAME) GetProcAddress(module, "gethostbyname");
    winSock.gethostname = (LPFN_GETHOSTNAME) GetProcAddress(module, "gethostname");
    winSock.getservbyname = (LPFN_GETSERVBYNAME) GetProcAddress(module, "getservbyname");
    winSock.getsockname = (LPFN_GETSOCKNAME) GetProcAddress(module, "getsockname");
    winSock.getpeername = (LPFN_GETPEERNAME) GetProcAddress(module, "getpeername");
    return TRUE;
}

#endif


static int erltk_init()
{
#ifdef __WIN32__
    WORD wVersionRequested;
    WSADATA wsaData;

    if (!lookup_functions())
	goto error;

    wVersionRequested = MAKEWORD(2,0);
    if ((*winSock.WSAStartup)(wVersionRequested, &wsaData) != 0)
	goto error;
    if ((LOBYTE(wsaData.wVersion) != 2) || (HIBYTE(wsaData.wVersion) != 0))
	goto error;
#endif
    LockInit(&erltk_lock);
    ThreadsInit();
    init_message_pool();

    if ((erltk_desc = malloc(sizeof(erltk_desc_t))) == NULL)
	goto error;
    erltk_desc->port = -1;
    erltk_desc->s     = INVALID_SOCKET;
    erltk_desc->s_tk  = INVALID_SOCKET;
    erltk_desc->event = INVALID_EVENT;
    erltk_desc->event_mask = 0;

    return 0;

error:
#ifdef __WIN32__
    (*winSock.WSACleanup)();
#endif
	/* The driver should be removed if not valid */
    remove_driver_entry(&tk_driver_entry);
    return -1;
}

static int erltk_finish()
{
    Lock(&erltk_lock);
    Unlock(&erltk_lock);
    LockDestroy(&erltk_lock);
    LockDestroy(&message_lock);
    if (erltk_desc != NULL)
	free(erltk_desc);
    return 0;
}

/* start the erltk main thread */
static long erltk_start(port, args)
long port; char* args;
{
    int len;
    erltk_desc_t* desc = erltk_desc;
    SOCKET pair[2];

    if (desc->port != -1)
	return -1;

    /* make a local copy of args (since we will patch it up) */
    len = strlen(args);
    if ((erltk_port_arg = (char*)malloc(len+1)) == NULL) {
	DPRINT("cound not allocate\n\r", 0);
	return -1;
    }
    strcpy(erltk_port_arg, args);
    if (sockpair(pair) < 0) {
	free(erltk_port_arg);
	return -1;
    }
    desc->s = pair[0];
    desc->s_tk = pair[1];

    /* Start TK thread */
    if (ThreadSpawn(ErlTk_Init, (void*) erltk_port_arg, &erltk_tid) < 0) {
	DPRINT("could not spawn thread\n\r", 0);
	sock_close(desc->s);
	sock_close(desc->s_tk);
	free(erltk_port_arg);
	return -1;
    }
    desc->port = port;
    desc->event = sock_create_event(desc);
    desc->event_mask = 0;
    sock_select(desc, (FD_READ | FD_CLOSE), 1);
#ifdef __WIN32__
    driver_select(desc->port, (int)desc->event, DO_READ, 1);
#endif
    return (long) desc;
}

/* stop the erltk main thread (close pipe, should terminate thread?) */
static int erltk_stop(desc)
erltk_desc_t* desc;
{
    erltk_mesg_t* mesg;
    int reason;

#ifdef __WIN32__
    driver_select(desc->port, (int)desc->event, DO_READ, 0);
#endif
    sock_select(desc, (FD_READ | FD_CLOSE), 0);
    sock_close_event(desc->event);
    desc->port = -1;

    sock_shutdown(desc->s, SD_SEND);
    while((mesg = ErlTk_RecvMsg(desc->s,&reason)) != NULL)
	free_message(mesg);
    sock_close(desc->s);

    /* ThreadJoin(erltk_tid, &result) */
    /* Since we cant destroy locks here tell start to doit next round */
    return 0;
}

/* put commands from erlang on tk input queue */
static int erltk_command(desc, buf, len)
erltk_desc_t* desc; char* buf; int len;
{
    erltk_mesg_t* mesg;

    /* allocate buffer */
    mesg = alloc_message(len);
    memcpy(mesg->data, buf, len);

    ErlTk_SendMsg(desc->s, mesg);
    return 1;
}

#ifdef __WIN32__

static int erltk_event(desc, event)
erltk_desc_t* desc; HANDLE event;
{
    WSANETWORKEVENTS netEv;
	int do_close = 1;

	TPRINT("erltk_event\n", 0);
    if ((*winSock.WSAEnumNetworkEvents)(desc->s, desc->event,&netEv) != 0)
		return -1;
    if (netEv.lNetworkEvents == 0)  /* NOTHING */
		return 0;

    if (netEv.lNetworkEvents & FD_READ) {
		do_close = 0;
		erltk_input(desc, event);
	}
    if (netEv.lNetworkEvents & FD_WRITE) {
		do_close = 0;
		erltk_output(desc, event);
	}

    if (do_close && (netEv.lNetworkEvents & FD_CLOSE)) {
		/* Handle close from TK */
		driver_failure(desc->port, 0);
		return -1;
    }
    return 0;
}

#endif

/* called when tk has formated messages to send or results */
static int erltk_input(desc, event)
erltk_desc_t* desc; HANDLE event;
{
    erltk_mesg_t* mesg;
    int reason;

    if ((mesg = ErlTk_RecvMsg(desc->s,&reason)) != NULL) {
	TPRINT2("erltk_input: message(%s) len = %d\n\r", 
		mesg->dynamic?"dynaic":"fixed", mesg->len);    
	driver_output(desc->port, (char*) mesg->data, mesg->len);
	free_message(mesg);
	return 1;
    }
    else if (reason == ERLTK_MESG_CLOSED) {
	DPRINT("erltk_input: tk closed\n\r", 0);
	driver_failure(desc->port, 0);
	return 0;
    }
    else {
	DPRINT("erltk_input: tk failure\n", 0);
	driver_failure(desc->port, -1);
	return -1;
    }
}

/* called when tk has processed a command and may recive more */
/* FIXME: when sender socket is full, we get a non-blocking
** condition. We must save the number of none sent events
** and send them when the socket is ready for write 
*/
static int erltk_output(desc, event)
erltk_desc_t* desc; HANDLE event;
{
    /* we may have to send more pipe wakeup's */
    /* messages shold have been put on the queue already */
    return 1;
}

